# Generated by roxygen2: do not edit by hand

export(GenerateNetwork)
export(GenerateSample)
export(plot_bbni)
export(plot_network)
export(plot_trace)
export(run_bbni)
importFrom(graphics,
  abline,
  legend,
  plot
)
importFrom(igraph,
  graph_from_adjacency_matrix,
  plot.igraph
)
importFrom(stats,
  rbinom,
  runif
)
importFrom(utils,
  setTxtProgressBar,
  txtProgressBar
)
