#' Yeast Cell-Cycle Gene Expression Data
#'
#' A binary empirical dataset with the gene expression states of 14 yeast
#' cell-cycle genes across 385 complete pooled experimental conditions. The
#' data were filtered to the 14 genes used by Han et al. (2014), columns with
#' missing values were removed, and each gene was binarized by its row mean.
#'
#' @format A numeric matrix with 14 rows (representing individual genes/nodes)
#' and 385 columns (representing independent samples or sequential time points).
#'
#' @source Han, S., Wong, R. K. W., Lee, T. C. M., Shen, L., Li, S.-Y. R., & Fan, X. (2014).
#' A Full Bayesian Approach for Boolean Genetic Network Inference. *PLOS ONE*, 9(12), e115806.
#' \doi{10.1371/journal.pone.0115806}
#'
#' @usage data(yeast_data)
"yeast_data"
